import base64
import codecs
import inspect
from pathlib import Path
from flask import request, jsonify, send_file
import lyrebird
from lyrebird import log
from lyrebird import application
from lyrebird.mock.context import make_fail_response, make_ok_response
from . import event_handler
from . import template_loader
from . import cache
from . import upload
from . import attachment
import traceback


logger = log.get_logger()


def template():
    if request.method == 'GET':
        '''
        Get all template list, all draft list, last selected template and last selected draft
        '''
        templates = template_loader.template_list()
        default_template = template_loader.get_default_template_path()

        draft_version = cache.check_draft_version()
        if draft_version < cache.DRAFT_VERSION_V_1_12_4:
            logger.info('Updating draft into v1.12.4')
            cache.update_selected_template()
            cache.update_all_draft_file(templates)

        last_selected_template = cache.get_selected_template()
        if not last_selected_template and default_template:
            last_selected_template = str(default_template)
        
        selected_template_index = None
        for index, template in enumerate(templates):
            if template['path'] == last_selected_template:
                selected_template_index = index
                break
        if selected_template_index is None:
            return application.make_ok_response(templates=templates)

        template_key = cache.get_filename(last_selected_template)
        drafts = cache.get_draft_list(template_key)
        last_selected_cache = cache.get_selected_cache()

        return application.make_ok_response(
            selected_template_index=selected_template_index,
            templates=templates,
            selected_cache=last_selected_cache,
            drafts=drafts
        )

    elif request.method == 'POST':
        '''
        Get template detail by path
        And save lastest_selected_template
        '''
        if not request.json or not request.json.get('path'):
            return application.make_fail_response('Missing required argument: path')

        template_path = request.json['path']
        draft_name = request.json.get('cache_name', '')
        draft_filename = cache.get_filename(template_path, draft_name)
        template_detail = cache.get(draft_filename)

        template = template_loader.get_template(template_path)
        if inspect.signature(template.form).parameters:
            template_detail = template.form({'cache': template_detail})
        else:
            template_detail = template.form()
        
        # 'template_loaded' fields will be ensured to have a uniform value by the script.
        if template_detail[0].get('template_loaded'):
            message = None
        else:
            message = 'Template loads failed! Please try again later.'

        cache.selected_template(template_path, draft_name)
        return application.make_ok_response(
            message=message,
            template_detail=template_detail
        )


def issue():
    issue_req = request.json
    template_info = issue_req['template']
    template = template_loader.get_template(template_info['path'])

    issue_data = issue_req['issue']
    attachments = issue_req['attachments']
    export_attachments = issue_req['exportAttachments']
    desc_img_id = issue_req['descImgId']

    all_bugit_message = ''

    # Export Snapshot/Json/... attachment files
    export_fail_attachments = []
    for attachment_item in export_attachments:
        if attachment_item.get('attachmentType') == 'lb':
            # Export Snapshot
            attachments.append(attachment.export_snapshot(attachment_item))
        else:
            # Export to specified file
            file_item = attachment.export_attachment_file(attachment_item)
            if 'id' in file_item:
                attachments.append(file_item)
            else:
                # export fail item do not have 'id' key
                export_fail_attachments.append(file_item['name'])

    # Set bugit script context
    context = {'issue': issue_data, 'attachments': attachments, 'desc_img_id': desc_img_id}
    # Set submit actions
    submit_action_functions = template.submit()
    # Do submit
    for action_function in submit_action_functions:
        try:
            bug_url = action_function(context)
            if bug_url:
                all_bugit_message += bug_url
        except Exception as e:
            if e.__class__.__name__ == 'BugitFormInputError':
                return application.make_fail_response(str(e))
            if e.__class__.__name__ == 'SubmitError':
                return application.make_fail_response(str(e.args[0]['message']),
                                                      template_detail=e.args[0]['template_detail'])
            error_message = traceback.format_exc()
            trace = "<br/>".join(error_message.split("\n"))
            lyrebird.report({
                'action': 'bugit.submit',
                'bugit': {
                    'status': 'failure',
                    'data': issue_data,
                    'error': error_message
                }
            })
            return application.make_fail_response(f'Script function "{action_function.__name__}" crash.<br/>{trace}')
    lyrebird.report({
        'action': 'bugit.submit',
        'bugit': {
            'status': 'success',
            'data': issue_data
        }
    })
    #Delete Snapshot
    attachment.remove_attach()
    return application.make_ok_response(
        message=f'Create issue success! {all_bugit_message}',
        export_fail_attachments=export_fail_attachments
    )


def take_screenshot(platform, device_id):
    """
    channel: {platform}.cmd
    """
    channel = platform+'.cmd'
    cmd = {}
    cmd['cmd'] = 'screenshot'
    cmd['device_id'] = [device_id]
    lyrebird.publish(channel.lower(), cmd)
    return application.make_ok_response(message=f'Take screenshot success!')

# Get available device from iOS&Android plugin
def device():
    if request.method == 'GET':
        all_devices = [
            *[_update_device_info(d, platform='Android')
              for d in event_handler.android_devices],
            *[_update_device_info(d, platform='iOS') for d in event_handler.ios_devices]
        ]
        return jsonify(all_devices)


def _update_device_info(device_info, platform="Android"):
    device_info['platform'] = platform
    return device_info


def attachments(attachment_id=None):
    if request.method == 'GET':
        if not attachment_id:
            attachment_list = list(event_handler.attachments.values())
            if not attachment_list:
                upload.reset_dir()
                attachment.remove_attach()
            return jsonify(attachment_list)
        else:
            attachment_item = event_handler.attachments.get(attachment_id)
            if attachment_item:
                return send_file(attachment_item['path'])
            else:
                return make_fail_response('Not found attachment')

    elif request.method == 'PUT':
        path = Path(event_handler.attachments[attachment_id].get('path'))
        name = event_handler.attachments[attachment_id].get('name')
        image_data = request.json.get('imageData')

        image_data = image_data.replace('data:image/png;base64,', '', 1)

        with codecs.open(str(path), 'wb') as f:
            image_data = base64.decodebytes(image_data.encode())
            f.write(image_data)
        return make_ok_response(message=f'Edited picture {name} has been saved')

    elif request.method == 'DELETE':
        attachment_item = event_handler.attachments.pop(attachment_id)
        upload.delete(attachment_item.get('path'))
        return make_ok_response()

    elif request.method == 'POST':
        if not attachment_id:
            # Upload attachment files
            if request.files:
                stream = request.files['file']
                if not stream:
                    return application.make_fail_response('Missing file data')

                upload.add(stream)
                return application.make_ok_response()
            return application.make_fail_response('No upload file found')

        # Rename
        attachment_item = event_handler.attachments.get(attachment_id)
        new_name = request.json.get('newName')
        try:
            attachment.rename(attachment_item, new_name)
            return application.make_ok_response()
        except FileNotFoundError as e:
            return application.make_fail_response(f'Rename attachment [{attachment_item.get("name")}] to [{new_name}] error: {e}')
            

def ui_cache(template_key):
    if request.method == 'GET':
        data = cache.get_draft_list(template_key)
        return application.make_ok_response(data=data)

    elif request.method == 'POST':
        template_path = request.json.get('template_path')
        if not template_path:
            return application.make_fail_response('Missing required argument: template_path')

        cache_name = request.json.get('cache_name')
        if not cache_name:
            return application.make_fail_response('Missing required argument: cache_name')

        template_detail = request.json.get('data')
        if not template_detail:
            return application.make_fail_response('Missing required argument: data')

        for field in template_detail:
            if 'extraMsg' in field:
                del field['extraMsg']

        draft_name = cache.get_filename(template_path, cache_name)
        cache.put(draft_name, template_detail)
        cache.put(f'{draft_name}{cache.DRAFT_INFO_FILE_SUFFIX}', {'cache_name': cache_name})
        cache.selected_template(template_path, cache_name)
        return application.make_ok_response()

    elif request.method == 'DELETE':
        template_path = request.json.get('template_path')
        if not template_path:
            return application.make_fail_response('Missing required argument: template_path')

        cache_name = request.json.get('cache_name')
        if not cache_name:
            return application.make_fail_response('Missing required argument: cache_name')

        draft_name = cache.get_filename(template_path, cache_name)
        cache.delete(draft_name)
        cache.delete(f'{draft_name}{cache.DRAFT_INFO_FILE_SUFFIX}')
        return application.make_ok_response()



import imp
import traceback
from hashlib import md5
from pathlib import Path
from lyrebird import application
from lyrebird import get_logger


logger = get_logger()

autoissue_ready = None
last_template = None

def get_workspace():
    bugit_workspace = application.config.get('bugit.workspace')

    if bugit_workspace and Path(bugit_workspace).exists():
        return Path(bugit_workspace)

    ROOT = application._cm.ROOT
    metadata_dir = ROOT/'downloads'/'lyrebird_bugit'
    metadata_dir.mkdir(parents=True, exist_ok=True)
    logger.warning(
        f'Bugit workspace not exists! Create on {metadata_dir._str}')

    return metadata_dir


def get_default_template_path():
    bugit_workspace = application.config.get('bugit.workspace')
    if not bugit_workspace:
        return
    
    bugit_default_template = application.config.get('bugit.default_template')
    if not bugit_default_template:
        return

    return Path(bugit_workspace) / Path(bugit_default_template)


def template_list():
    template_list = []
    for template_file in get_workspace().iterdir():
        if not template_file.name.endswith('.py'):
            continue
        try:
            logger.debug(f'Load template {template_file}')
            template = imp.load_source(template_file.stem, str(template_file))
            template_check(template)
            relative_template_file_path = str(template_file).replace(str(Path.home()), '~')
            template_key = md5(relative_template_file_path.encode()).hexdigest()
            template_list.append({
                'name': template.name,
                'path': str(template_file),
                'id': template_key
            })
            del template
        except Exception:
            logger.error(
                f'Load bug template failed:\nBad template: {template_file}\n{traceback.format_exc()}')
    return template_list


def template_check(template):
    assert hasattr(template, 'name'), "BugIt template should have name attr"
    assert hasattr(template, 'form'), "BugIt template should have form attr"
    assert callable(template.form), "BugIt template should have form function"
    assert hasattr(
        template, 'submit'), "BugIt template should have submit attr"
    assert callable(
        template.submit), "BugIt template should have submit function"


def default_template_check(template_path):
    global autoissue_ready

    if not template_path:
        logger.error('Default template path is not configured.')
        autoissue_ready = False
        return

    if not template_path.exists():
        logger.error('Default template path is not existed.')
        autoissue_ready = False
        return
    
    template = get_template(template_path)
    if not (hasattr(template, 'auto_issue') and callable(template.auto_issue)):
        logger.error('Default template should have auto_issue function.')
        autoissue_ready = False
        return
    
    autoissue_ready = True


def get_template(file_path):
    global last_template

    if not last_template or last_template.__file__ != str(file_path):
        last_template = imp.load_source(Path(file_path).stem, str(file_path))

    return last_template


def notice_handler(msg):
    default_template_path = get_default_template_path()

    if autoissue_ready is None:
        default_template_check(default_template_path)
    
    if autoissue_ready == False:
        return
    
    # Filter out messages with invalid types
    if not isinstance(msg, dict):
        return
    
    template = get_template(default_template_path)
    template.auto_issue(msg)



import json
import codecs
import lyrebird
import os
from hashlib import md5
from pathlib import Path
from packaging import version


logger = lyrebird.log.get_logger()

BUGIT_STORAGE = lyrebird.get_plugin_storage()
CACHE_ROOT = Path(BUGIT_STORAGE)/'cache'
LAST_SELECT_TEMPLATE_FILENAME = 'selected_template'
BUGIT_SYSTEM_FILENAME = [LAST_SELECT_TEMPLATE_FILENAME]
DRAFT_INFO_FILE_SUFFIX = '.info'
DEFAULT_DRAFT_NAME = 'Default'
DRAFT_VERSION_V_1_8_0 = version.parse('1.8.0')
DRAFT_VERSION_V_1_12_4 = version.parse('1.12.4')


def get(template_key):
    return _read_cache_from_file(template_key)


def put(template_key, data):
    _save_cache_to_file(template_key, data)


def delete(template_key):
    _delete_cache_from_file(template_key)


def get_filename(abs_template_path, draft_name=''):
    relative_template_path = _change_home(abs_template_path, str(Path.home()), '~')
    template_key = md5(relative_template_path.encode()).hexdigest()
    if not draft_name:
        return f'{template_key}'

    draft_key = md5(draft_name.encode()).hexdigest()
    return f'{template_key}_{draft_key}'


def selected_template(abs_template_path, cache_name):
    relative_template_path = _change_home(abs_template_path, str(Path.home()), '~')
    _save_cache_to_file(LAST_SELECT_TEMPLATE_FILENAME, 
                        {'path': relative_template_path, 'cache_name': cache_name, 'draft_version': '1.12.4'})


def get_selected_template():
    _cache = _read_cache_from_file(LAST_SELECT_TEMPLATE_FILENAME)
    if not _cache:
        return None
    selected_template_path = _cache.get('path')
    return str(Path(selected_template_path).expanduser())


def get_selected_cache():
    _cache = _read_cache_from_file(LAST_SELECT_TEMPLATE_FILENAME)
    if not _cache:
        return None
    return _cache.get('cache_name')


def get_draft_list(template_key):
    draft_list = []
    for draft_file in CACHE_ROOT.iterdir():
        if draft_file.suffix != DRAFT_INFO_FILE_SUFFIX:
            continue
        if template_key not in str(draft_file):
            continue

        try:
            with codecs.open(draft_file, 'r') as f:
                cache_name = json.load(f).get('cache_name')
                draft_list.append({'cacheName': cache_name})
        except Exception:
            logger.error(f'Load BugIt cache fail! {str(draft_file)}')

    draft_list.sort(key=lambda x: x['cacheName'])
    return draft_list


def check_draft_version():
    _cache = _read_cache_from_file(LAST_SELECT_TEMPLATE_FILENAME)
    if not _cache:
        return DRAFT_VERSION_V_1_12_4
    if 'draft_version' in _cache:
        return DRAFT_VERSION_V_1_12_4
    return DRAFT_VERSION_V_1_8_0


def update_selected_template():
    _cache = _read_cache_from_file(LAST_SELECT_TEMPLATE_FILENAME)
    template_path = _change_home(_cache['path'], str(Path.home()), '~')  # for macOS and winOS
    template_path = _change_home(_cache['path'], '/root', '~')  # for Linux OS
    _cache['path'] = template_path
    if 'cache_name' not in _cache:
        _cache['cache_name'] = DEFAULT_DRAFT_NAME
    _cache['draft_version'] = '1.12.4'
    _save_cache_to_file(LAST_SELECT_TEMPLATE_FILENAME, _cache)


def generate_template_keys(templates):
    if 'HOST_HOME' in os.environ:
        other_home_path = os.environ['HOST_HOME']
    else:
        other_home_path = '/root'
    
    template_keys_map = {}
    for template in templates:
        new_template_key = template['id']
        template_key = str(md5(template['path'].encode()).hexdigest())
        template_keys_map[template_key] = new_template_key
        other_absolute_path = _change_home(template['path'], str(Path.home()), other_home_path)
        other_template_key = str(md5(other_absolute_path.encode()).hexdigest())
        template_keys_map[other_template_key] = new_template_key
    return template_keys_map


def update_all_draft_file(templates):
    template_keys_map = generate_template_keys(templates)

    for cache_file in CACHE_ROOT.iterdir():
        if cache_file.name.startswith('.'):
            continue
        if cache_file.name in BUGIT_SYSTEM_FILENAME:
            continue
        if cache_file.name.endswith(DRAFT_INFO_FILE_SUFFIX):
            continue

        cache_name_parts = cache_file.name.split('_')
        if len(cache_name_parts) < 2:
            continue
        old_template_key = cache_name_parts[0]
        cache_key = cache_name_parts[1]
        if old_template_key not in template_keys_map:
            continue
        
        new_template_key = template_keys_map[old_template_key]
        file_name = f'{new_template_key}_{cache_key}'
        detail_path = CACHE_ROOT / file_name
        info_path = CACHE_ROOT / f'{cache_file}.info'
        new_info_path = CACHE_ROOT / f'{file_name}.info'
        if detail_path.exists():
            try:
                with codecs.open(str(info_path), 'r') as f:
                    cache_name = json.load(f)['cache_name']
            except Exception:
                logger.error(f'Update cache from v1.8.0 to v1.12.4 error! \
                             Cannot read the info file: {str(cache_file)}')
                continue
            cache_name = f'{cache_name}_copy'
            try:
                with open(info_path, 'w') as f:
                    f.write(json.dumps({'cache_name': cache_name}, ensure_ascii=False))
            except Exception:
                logger.error(f'Update cache from v1.8.0 to v1.12.4 error! \
                             Cannot write the info file: {str(cache_file)}')
                continue
            cache_key = str(md5(cache_name.encode()).hexdigest())
            file_name = f'{new_template_key}_{cache_key}'
            detail_path = CACHE_ROOT / file_name
            new_info_path = CACHE_ROOT / f'{file_name}.info'
        cache_file.rename(detail_path)
        info_path.rename(new_info_path)


def _check_dir():
    if not CACHE_ROOT.exists():
        CACHE_ROOT.mkdir(parents=True, exist_ok=True)


def _save_cache_to_file(name, data):
    _check_dir()
    with codecs.open(str(CACHE_ROOT / name), 'w') as f:
        f.write(json.dumps(data, ensure_ascii=False))


def _read_cache_from_file(name):
    _check_dir()
    target_file = CACHE_ROOT / name
    if not target_file.exists():
        return None
    try:
        with codecs.open(str(target_file), 'r') as f:
            return json.load(f)
    except Exception:
        return None

def _delete_cache_from_file(name):
    target_file = CACHE_ROOT / name
    if target_file.exists() and target_file.is_file():
        target_file.unlink()


def _change_home(original_path, old_home, new_home):
    """
    Change the home part of the original path:
    from an old OS to a new OS, or from absolute to relative (when new_home = '~')
    """
    _path = original_path.replace(old_home, new_home)
    return _path



import lyrebird
from . import template_loader
from uuid import uuid4
from collections import OrderedDict
from lyrebird import get_logger
from lyrebird import application


logger = get_logger()


android_devices = []
ios_devices = []
attachments = OrderedDict()


def on_android_device(msg):
    logger.debug(f'PluginHandler: [Android.device] {msg}')
    global android_devices
    android_devices = msg
    lyrebird.emit('devices')


def on_ios_device(msg):
    logger.debug(f'PluginHandler: [iOS.device] {msg}')
    global ios_devices
    ios_devices = msg
    lyrebird.emit('devices')


def on_android_screenshot(msg):
    if len(msg) == 0:
        return
    for item in msg:
        attachment_id = str(uuid4())
        attachments[attachment_id] = {'id': attachment_id,
                                      'name': item['screenshot']['name'],
                                      'path': item['screenshot']['path']}
    lyrebird.emit('attachments')


def on_ios_screenshot(msg):
    if len(msg) == 0:
        return
    for item in msg:
        attachment_id = str(uuid4())
        attachments[attachment_id] = {'id': attachment_id,
                                      'name': item['screenshot']['name'],
                                      'path': item['screenshot']['path']}
    lyrebird.emit('attachments')


def on_upload_files(msg):
    if len(msg) == 0:
        return
    for item in msg:
        attachment_id = str(uuid4())
        attachments[attachment_id] = {'id': attachment_id,
                                      'name': item['upload_file']['name'],
                                      'path': item['upload_file']['path']}
    lyrebird.emit('attachments')


def on_notice(msg):
    sender_file = msg.get('sender', {}).get('file', '')
    autoissue_checker = application.config.get('event.notice.autoissue.checker', [])
    if sender_file in autoissue_checker:
        template_loader.notice_handler(msg)
